<?php
/*
* index.php
*
* Copyright (c) 2014, Swapnajit Mitra. All rights reserved.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
* MA 02110-1301 USA
*/

/*
*  This is a PHP implementation of David H Bailey's original C 
*  program piqpr8.c to generate HEX digits of pi from a given location.
*/
$NHX = 16;
$eps = 1e-17;
$ntp = 25;
/* Static tp[ntp]; */
$tp1 = 0; /* static */

$id = rand(0,100000); /* id is the digit position. Digits generated follow immediately after id. */
$nhx = 10;

if (isset($_GET["s"])) {
   if (is_numeric($_GET["s"])) { 
      $id = htmlspecialchars($_GET["s"]);
   } else { 
      /*
      echo "ERROR: Parameter s must be an integer";
      */
      exit; 
   }
}

if (isset($_GET["n"])) {
   if (is_numeric($_GET["n"])) { 
      $nhx = htmlspecialchars($_GET["n"]);
   } else { 
      /*
      echo "ERROR: Parameter n must be an integer";
      */
      exit; 
   }
   if ($nhx > 1000) { 
      $nhx = 1000;
   }
}

header('Content-type: application/xml;charset=utf-8');
$xml = new SimpleXMLElement("<?xml version=\"1.0\" encoding=\"utf-8\" ?><Random-hex-pi></Random-hex-pi>");
$xml->addChild('start-position-in-decimal', "$id");
$xml->addChild('num-hex-digits', "$nhx");

$value = '';
do { 
   $s1 = series (1, $id);
   $s2 = series (4, $id);
   $s3 = series (5, $id);
   $s4 = series (6, $id);
   $pid = 4.0 * $s1 - 2.0 * $s2 - $s3 - $s4;
   $pid = $pid - (int) $pid + 1.0;
   $chx = ihex ($pid, $NHX);
   
   $value .= sprintf ("%10.10s", $chx);

   $nhx -= 10;
   $id += 10;
} while ($nhx>0); 

$xml->addChild('value', "$value");
print($xml->asXML());

function ihex ($x, $nhx) { 

/*  This returns, in chx, the first nhx hex digits of the fraction of x. */

   $chx = "abcd";
   $hx = "0123456789ABCDEF";
   $y = abs ($x);

   for ($i=0; $i < $nhx; $i++) { 
      $y = 16.0 * ($y - floor($y));
      $chx = substr_replace ($chx, substr ($hx, (int) $y, 1), $i, 1); 
   }
   return $chx;
}

function series ($m, $id) {

/*  This routine evaluates the series  sum_k 16^(id-k)/(8*k+m) 
    using the modular exponentiation technique. */

  global $eps;
  $s = 0.0;

/*  Sum the series up to id. */

  for ($k = 0; $k < $id; $k++){
    $ak = 8 * $k + $m;
    $p = $id - $k;
    $t = expm ($p, $ak);
    $s = $s + $t / $ak;
    $s = $s - (int) $s;
  }

/*  Compute a few terms where k >= id. */

  for ($k = $id; $k <= $id + 100; $k++){
    $ak = 8 * $k + $m;
    $t = pow (16, ($id - $k)) / $ak;
    if ($t < $eps) break;
    $s = $s + $t;
    $s = $s - (int) $s;
  }
  return $s;
}

function expm ($p, $ak) { 

/*  expm = 16^p mod ak.  This routine uses the left-to-right binary 
    exponentiation scheme. */

/*  If this is the first call to expm, fill the power of two table tp. */
global $tp1; 
global $tp; 
global $ntp;

  if ($tp1 == 0) {
    $tp1 = 1;
    $tp[0] = 1.0;

    for ($i = 1; $i < $ntp; $i++) {
       $tp[$i] = 2.0 * $tp[$i-1];
    }   
  }

  if ($ak == 1.0) return 0.0;

/*  Find the greatest power of two less than or equal to p. */

  for ($i = 0; $i < $ntp; $i++) {
     if ($tp[$i] > $p) {
        break;
     }
  }	

  $pt = $tp[($i-1)];
  $p1 = $p;
  $r = 1.0;

/*  Perform binary exponentiation algorithm modulo ak. */

  for ($j = 1; $j <= $i; $j++){
    if ($p1 >= $pt){
      $r = 16.0 * $r;
      $r = $r - (int) ($r / $ak) * $ak;
      $p1 = $p1 - $pt;
    }
    $pt = 0.5 * $pt;
    if ($pt >= 1.0){
      $r = $r * $r;
      $r = $r - (int) ($r / $ak) * $ak;
    }
  }

  return $r;
}
?>
